"""EVCC Devices."""

import logging
import uuid

from energy_assistant import Optimizer
from energy_assistant.constants import ROOT_LOGGER_NAME
from energy_assistant.devices.analysis import FloatDataBuffer
from energy_assistant.devices.homeassistant import HOMEASSISTANT_CHANNEL, HomeassistantState, UnavailableUnitError

from . import (
    LoadInfo,
    OnOffState,
    PowerModes,
    SessionStorage,
    State,
    StateId,
    StatesRepository,
)
from .config import get_config_param
from .device import DeviceWithState

LOGGER = logging.getLogger(ROOT_LOGGER_NAME)


class HomeassistantEvccDevice(DeviceWithState):
    """Evcc load points as devices."""

    def __init__(self, device_id: uuid.UUID, session_storage: SessionStorage) -> None:
        """Create an evcc device."""
        super().__init__(device_id, session_storage)
        self._evcc_topic: str = ""
        self._loadpoint_name: str = ""
        self._is_continous: bool = True
        self._nominal_power: float | None = None
        self._state = OnOffState.UNKNOWN
        self._power: State | None = None
        self._consumed_energy: State | None = None
        self._mode: State | None = None
        self._vehicle_soc: State | None = None
        self._max_current: State | None = None
        self._is_connected: State | None = None
        self._supported_power_modes = {
            PowerModes.DEVICE_CONTROLLED,
            PowerModes.OFF,
            PowerModes.PV,
            PowerModes.MIN_PV,
            PowerModes.FAST,
            PowerModes.OPTIMIZED,
        }
        self._utility_meter = self.add_utility_meter("energy")

    def configure(self, config: dict) -> None:
        """Load the device configuration from the provided data."""
        super().configure(config)
        LOGGER.debug(f"Configure EVCC device {self._name} with config: {config}")
        self._evcc_topic = get_config_param(config, "evcc_topic")
        self._loadpoint_name = get_config_param(config, "load_point_name")
        self._is_continous = bool(config.get("continuous", True))
        self._nominal_power = config.get("nominal_power")

    @property
    def type(self) -> str:
        """The device type."""
        return "evcc"

    def get_device_topic_id(self, name: str, entity_type: str = "sensor") -> StateId:
        """Get the id of a topic of this load point."""
        return StateId(
            id=f"{entity_type}.evcc_{self._loadpoint_name}_{name}",
            channel=HOMEASSISTANT_CHANNEL,
        )

    async def update_state(self, state_repository: StatesRepository, self_sufficiency: float) -> None:
        """Update the state of the Stiebel Eltron device."""
        old_state = self.state == OnOffState.ON
        charging = state_repository.get_state(self.get_device_topic_id("charging", "binary_sensor"))
        if charging is not None:
            self._state = OnOffState.ON if charging.value == "true" else OnOffState.OFF
        else:
            self._state = OnOffState.UNKNOWN
        new_state = self.state == OnOffState.ON

        self._consumed_energy = state_repository.get_state(self.get_device_topic_id("charge_total_import"))
        if self.consumed_energy == 0:
            state = state_repository.get_state(self.get_device_topic_id("session_energy"))
            if state is not None:
                self._consumed_energy = self._utility_meter.update_energy_state(state)
        self._consumed_solar_energy.add_measurement(self.consumed_energy, self_sufficiency)
        self._power = state_repository.get_state(self.get_device_topic_id("charge_power"))
        self._mode = state_repository.get_state(self.get_device_topic_id("mode", "select"))
        self._vehicle_soc = state_repository.get_state(self.get_device_topic_id("vehicle_soc"))
        self._vehicle_capacity = state_repository.get_state(self.get_device_topic_id("vehicle_capacity"))
        self._max_current = state_repository.get_state(self.get_device_topic_id("max_current", "select"))
        self._is_connected = state_repository.get_state(self.get_device_topic_id("connected", "binary_sensor"))

        if self._energy_snapshot is None:
            self.set_snapshot(self.consumed_solar_energy, self.consumed_energy)

        await self.update_session(old_state, new_state, "EVCC")

    async def update_power_consumption(
        self,
        state_repository: StatesRepository,
        optimizer: Optimizer,
        grid_exported_power_data: FloatDataBuffer,
    ) -> None:
        """Update the device based on the current pv availability."""
        if self.power_mode != PowerModes.DEVICE_CONTROLLED:
            new_state = ""
            if self.power_mode == PowerModes.OFF:
                new_state = "off"
            elif self.power_mode == PowerModes.PV:
                new_state = "pv"
            elif self.power_mode == PowerModes.MIN_PV:
                new_state = "minpv"
            elif self.power_mode == PowerModes.FAST:
                new_state = "now"
            elif self.power_mode == PowerModes.OPTIMIZED:
                # TODO: Implement Optimized with emhass
                new_state = "pv"
            if new_state != self._mode:
                state_repository.set_state(self.get_device_topic_id("mode"), new_state)

    @property
    def consumed_energy(self) -> float:
        """Consumed energy in kWh."""
        return self._consumed_energy.numeric_value if self._consumed_energy else 0.0

    @property
    def state(self) -> str:
        """The state of the device. The state is `on` in case the device is loading."""
        return self._state

    @property
    def power(self) -> float:
        """Current power consumption of the device."""
        if self._power is None:
            return 0.0
        if isinstance(self._power, HomeassistantState):
            unit = str(getattr(self._power, "unit", ""))
            if unit == "kW":
                return self._power.numeric_value * 1000
            if unit == "W":
                return self._power.numeric_value
            raise UnavailableUnitError(unit)
        return self._power.numeric_value

    @property
    def mode(self) -> str:
        """Current PV mode of the device."""
        return self._mode.value if self._mode else "unknown"

    @property
    def vehicle_soc(self) -> float:
        """State of Charge of the connected vehicle."""
        return self._vehicle_soc.numeric_value if self._vehicle_soc else 0.0

    @property
    def vehicle_capacity(self) -> float:
        """Capacity of the connected vehicle."""
        return self._vehicle_capacity.numeric_value if self._vehicle_capacity else 0.0

    @property
    def icon(self) -> str:
        """Icon of the device."""
        return "mdi-car-electric"

    @property
    def available(self) -> bool:
        """Check if the device is available."""
        return True

    def restore_state(self, consumed_solar_energy: float, consumed_energy: float) -> None:
        """Restore the previously stored state."""
        super().restore_state(consumed_solar_energy, consumed_energy)
        self._consumed_energy = State(self.get_device_topic_id("chargeTotalImport").id, str(consumed_energy))

    @property
    def attributes(self) -> dict[str, str]:
        """Get the attributes of the device for the UI."""
        result: dict[str, str] = {
            **super().attributes,
            "pv_mode": self.mode,
        }
        if self._vehicle_soc is not None:
            result["vehicle_soc"] = f"{round(self.vehicle_soc)} %"
        return result

    def get_load_info(self) -> LoadInfo | None:
        """Get the current deferrable load info."""
        if (
            self.state == OnOffState.ON
            and self._is_connected is not None
            and self._max_current is not None
            and self._is_connected.value == "true"
        ):
            remaining_energy = (1 - self.vehicle_soc / 100) * self.vehicle_capacity * 1000
            if remaining_energy > 0:
                power: float = (
                    self._nominal_power if self._nominal_power is not None else (self._max_current.numeric_value * 230)
                )  # TODO: Multiply with active phases
                return LoadInfo(
                    device_id=self.id,
                    nominal_power=power,
                    duration=remaining_energy / power * 3600,
                    is_continous=self._is_continous,
                    is_deferrable=self.power_mode == PowerModes.OPTIMIZED,
                )
        return None



"""Schemas for home measurement api."""

from pydantic import BaseModel


class ConfigModel(BaseModel):
    """Schema for the configuration."""

    config: dict


class ReadConfigResponse(ConfigModel):
    """API Response for reading the configuration."""


class ReadDeviceConfigResponse(ConfigModel):
    """API Response for reading the device configuration."""


class VersionModel(BaseModel):
    """Schema for the version."""

    version: str
    ui_version: str


class ReadVersionResponse(VersionModel):
    """API Response for reading the version."""


class DeviceControlModel(BaseModel):
    """Schema for the device control."""

    disable_device_control: bool


class ReadDeviceControlResponse(DeviceControlModel):
    """API Response for reading the device control."""



"""Generic Home Assistant device."""

import logging
import uuid
from datetime import UTC, datetime

from energy_assistant import Optimizer
from energy_assistant.constants import (
    DEFAULT_NOMINAL_DURATION,
    DEFAULT_NOMINAL_POWER,
    POWER_HYSTERESIS,
    ROOT_LOGGER_NAME,
)
from energy_assistant.devices import (
    LoadInfo,
    OnOffState,
    PowerModes,
    SessionStorage,
    State,
    StateId,
    StatesRepository,
    assign_if_available,
)
from energy_assistant.devices.analysis import FloatDataBuffer, OnOffDataBuffer
from energy_assistant.devices.config import get_config_param
from energy_assistant.devices.device import DeviceWithState
from energy_assistant.devices.homeassistant import HOMEASSISTANT_CHANNEL, HomeassistantState
from energy_assistant.devices.registry import DeviceType, DeviceTypeRegistry
from energy_assistant.devices.state_value import StateValue
from energy_assistant.storage.config import DeviceConfigMissingParameterError

LOGGER = logging.getLogger(ROOT_LOGGER_NAME)


class HomeassistantDevice(DeviceWithState):
    """A generic Home Assistant device."""

    def __init__(
        self,
        device_id: uuid.UUID,
        session_storage: SessionStorage,
        device_type_registry: DeviceTypeRegistry,
    ) -> None:
        """Create a generic Home Assistant device."""
        super().__init__(device_id, session_storage)
        self._device_type_registry = device_type_registry
        self._nominal_power: float | None = None
        self._nominal_duration: float | None = None
        self._is_constant: bool | None = None

        self._power_entity_id: str = ""
        self._power: State | None = None
        self._consumed_energy: State | None = None
        self._output_state: State | None = None
        self._output_id: str | None = None
        self._icon: str = "mdi-home"
        self._power_data = FloatDataBuffer()
        self._output_states = OnOffDataBuffer()

        self._state: str = OnOffState.UNKNOWN
        self._consumed_energy_value: StateValue | None = None
        self._device_type: DeviceType | None = None

        self._max_on_per_day: float = 24 * 60 * 60  # seconds
        self._min_on_duration: float = 0.0
        self._switch_off_delay: float = 0.0
        self._switch_on_delay: float = 0.0

    def configure(self, config: dict) -> None:
        """Load the device configuration from the provided data."""
        super().configure(config)
        self._power_entity_id = get_config_param(config, "power")
        energy_config = config.get("energy")
        if energy_config is not None:
            self._consumed_energy_value = StateValue(energy_config)
        else:
            msg = "energy"
            raise DeviceConfigMissingParameterError(msg)

        energy_scale: float | None = config.get("energy_scale")
        if energy_scale is not None:
            self._consumed_energy_value.set_scale(energy_scale)
            LOGGER.warning(
                f"Homeassistant device with id {self.id} is configured with energy_scale. This is deprecated and will no longer be supported.",
            )

        self._output_id = config.get("output")
        self._icon = str(config.get("icon", "mdi-home"))

        if self._output_id is not None:
            self._supported_power_modes.add(PowerModes.PV)
            self._supported_power_modes.add(PowerModes.OPTIMIZED)

        manufacturer = config.get("manufacturer")
        model = config.get("model")

        if model is not None and manufacturer is not None:
            self._device_type = self._device_type_registry.get_device_type(manufacturer, model)
            self._nominal_power = config.get(
                "nominal_power",
                self._device_type.nominal_power if self._device_type else None,
            )
            self._nominal_duration = config.get(
                "nominal_duration",
                self._device_type.nominal_duration if self._device_type else None,
            )
            self._is_constant = config.get("constant", self._device_type.constant if self._device_type else None)
        else:
            nominal_power = config.get("nominal_power")
            if nominal_power is not None:
                self._nominal_power = float(nominal_power)
            self._nominal_duration = config.get("nominal_duration")
            self._is_constant = config.get("constant")
        self._max_on_per_day = config.get("max_on_per_day", 24 * 60 * 60)  # seconds
        self._min_on_duration = config.get("min_on_duration", 60.0)  # seconds
        self._switch_off_delay = config.get("switch_off_delay", 300.0)  # seconds
        self._switch_on_delay = config.get("switch_on_delay", 300.0)  # seconds
        if self._device_type is None and (state_detection := config.get("state", {})):
            state_on = state_detection.get("state_on", {})
            state_off = state_detection.get("state_off", {})
            self._device_type = DeviceType(
                str(config.get("icon", "mdi:lightning-bolt")),
                self._nominal_power if self._nominal_power is not None else DEFAULT_NOMINAL_POWER,
                (self._nominal_duration if self._nominal_duration is not None else DEFAULT_NOMINAL_DURATION),
                self._is_constant if self._is_constant is not None else False,
                state_on.get("threshold", 2),
                state_off.get("threshold", 0),
                state_off.get("upper", 2.0),
                state_off.get("lower", 0),
                state_off.get("for", 0),
                state_off.get("trailing_zeros_for", 10),
            )

    @property
    def type(self) -> str:
        """The device type."""
        return "homeassistant"

    async def update_state(self, state_repository: StatesRepository, self_sufficiency: float) -> None:
        """Update the own state from the states of a StatesRepository."""
        if self._output_id is not None:
            self._output_state = assign_if_available(self._output_state, state_repository.get_state(self._output_id))
        else:
            self._output_state = None
        self._power = assign_if_available(self._power, state_repository.get_state(self._power_entity_id))
        self._consumed_energy = assign_if_available(
            self._consumed_energy,
            self._consumed_energy_value.evaluate(state_repository) if self._consumed_energy_value is not None else None,
        )
        self._consumed_solar_energy.add_measurement(self.consumed_energy, self_sufficiency)
        if self._energy_snapshot is None:
            self.set_snapshot(self.consumed_solar_energy, self.consumed_energy)

        if self.has_state:
            old_state = self.state == OnOffState.ON
            if self._device_type is not None:
                self._power_data.add_data_point(self.power)
                if self.state != OnOffState.ON and self.power > self._device_type.state_on_threshold:
                    self._state = OnOffState.ON
                elif self.state != OnOffState.OFF:
                    if self.state == OnOffState.ON and self.power <= self._device_type.state_off_threshold:
                        is_between = self._device_type.state_off_for > 0 and self._power_data.is_between(
                            self._device_type.state_off_lower,
                            self._device_type.state_off_upper,
                            self._device_type.state_off_for,
                            without_trailing_zeros=True,
                        )
                        max_value = self._device_type.trailing_zeros_for > 0 and self._power_data.get_max_for(
                            self._device_type.trailing_zeros_for,
                        )
                        if is_between or max_value <= self._device_type.state_off_threshold:
                            self._state = OnOffState.OFF
                    elif self.state == OnOffState.UNKNOWN:
                        self._state = OnOffState.OFF
            new_state = self.state == OnOffState.ON
            await super().update_session(old_state, new_state, "Power State Device")

    async def update_power_consumption(
        self,
        state_repository: StatesRepository,
        optimizer: Optimizer,
        grid_exported_power_data: FloatDataBuffer,
    ) -> None:
        """Update the device based on the current pv availability."""
        if self._output_id is None:
            return
        state: bool = self._output_state.value == "on" if self._output_state is not None else False
        new_state = state
        if self.power_mode == PowerModes.PV:
            midnight = datetime.now(UTC).replace(hour=0, minute=0, second=0, microsecond=0)
            if state:
                max_grid_power = grid_exported_power_data.get_max_for(self._switch_off_delay)
                if (
                    max_grid_power < self.nominal_power * (1 - POWER_HYSTERESIS)
                    and self._output_states.duration_in_state(True).total_seconds() > self._min_on_duration
                ) or self._output_states.total_duration_in_state_since(
                    True, midnight
                ).total_seconds() > self._max_on_per_day:
                    new_state = False
            else:
                # If the device is off, we check if the average power is below the nominal power
                # to turn it on again.
                min_grid_power = grid_exported_power_data.get_min_for(self._switch_on_delay)
                if (
                    min_grid_power > self.nominal_power * (1 + POWER_HYSTERESIS)
                    and self._output_states.total_duration_in_state_since(True, midnight).total_seconds()
                    < self._max_on_per_day
                ):
                    new_state = True

        elif self.power_mode == PowerModes.OPTIMIZED:
            power = optimizer.get_optimized_power(self._id)
            new_state = power > 0
        if state != new_state:
            state_repository.set_state(
                StateId(
                    id=self._output_id,
                    channel=HOMEASSISTANT_CHANNEL,
                ),
                "on" if new_state else "off",
            )
            self._output_states.add_data_point(new_state)

    @property
    def consumed_energy(self) -> float:
        """The consumed energy of the device."""
        return self._consumed_energy.numeric_value if self._consumed_energy is not None else 0.0

    @property
    def icon(self) -> str:
        """The icon of the device."""
        return self._icon

    @property
    def power(self) -> float:
        """The current power used by the device."""
        return self._power.numeric_value if self._power else 0.0

    @property
    def available(self) -> bool:
        """Is the device available?."""
        return (
            self._consumed_energy is not None
            and self._consumed_energy.available
            and self._power is not None
            and self._power.available
        )

    @property
    def nominal_power(self) -> float:
        """The nominal power of the device."""
        return self._nominal_power if self._nominal_power is not None else DEFAULT_NOMINAL_POWER

    def restore_state(self, consumed_solar_energy: float, consumed_energy: float) -> None:
        """Restore a previously stored state."""
        super().restore_state(consumed_solar_energy, consumed_energy)
        self._consumed_energy = HomeassistantState("", str(consumed_energy))

    @property
    def state(self) -> str:
        """The state of the device. The state is `on` in case the device is running."""
        return self._state

    @property
    def has_state(self) -> bool:
        """Has this device a state."""
        return self._device_type is not None

    def get_load_info(self) -> LoadInfo | None:
        """Get the current deferrable load info."""
        if self._nominal_power is not None and self._nominal_duration is not None:
            if self.power_mode == PowerModes.OPTIMIZED:
                return LoadInfo(
                    device_id=self.id,
                    nominal_power=self._nominal_power,
                    duration=self._nominal_duration,
                    is_continous=False,
                    is_constant=self._is_constant if self._is_constant is not None else False,
                    is_deferrable=True,
                )

            if self.state == OnOffState.ON:
                return LoadInfo(
                    device_id=self.id,
                    nominal_power=self._nominal_power,
                    duration=self._nominal_duration - self.session_duration,
                    is_continous=False,
                    is_constant=self._is_constant if self._is_constant is not None else False,
                    is_deferrable=False,
                )

        return None



"""Views for home measurement API."""

import uuid
from typing import Annotated

from fastapi import APIRouter, Depends, HTTPException, Path, Request

from .schema import (
    ConfigModel,
    DeviceControlModel,
    ReadConfigResponse,
    ReadDeviceConfigResponse,
    ReadDeviceControlResponse,
    ReadVersionResponse,
    VersionModel,
)
from .use_cases import (
    ReadConfiguration,
    ReadDeviceConfiguration,
    ReadDeviceControl,
    ReadVersion,
    WriteDeviceConfiguration,
    WriteDeviceControl,
)

router = APIRouter(prefix="/config")


@router.get("", response_model=ReadConfigResponse)
async def read_configuration(
    request: Request,
    use_case: Annotated[ReadConfiguration, Depends(ReadConfiguration)],
) -> ConfigModel:
    """Rest end point for read all devices."""
    energy_assistant = request.app.energy_assistant if hasattr(request.app, "energy_assistant") else None
    if energy_assistant is None:
        raise HTTPException(status_code=500)
    return await use_case.execute(energy_assistant.config)


@router.get("/version", response_model=ReadVersionResponse)
async def read_version(
    request: Request,
    use_case: Annotated[ReadVersion, Depends(ReadVersion)],
) -> VersionModel:
    """Rest end point for read all devices."""

    return await use_case.execute()


@router.get("/device_control", response_model=ReadDeviceControlResponse)
async def read_device_control(
    request: Request,
    use_case: Annotated[ReadDeviceControl, Depends(ReadDeviceControl)],
) -> DeviceControlModel:
    """Rest end point for read all devices."""
    energy_assistant = request.app.energy_assistant if hasattr(request.app, "energy_assistant") else None
    if energy_assistant is None:
        raise HTTPException(status_code=500)
    return await use_case.execute(energy_assistant.home)


@router.put("/device_control", response_model=ReadDeviceControlResponse)
async def write_device_control(
    request: Request,
    disable_device_control: bool,
    use_case: Annotated[WriteDeviceControl, Depends(WriteDeviceControl)],
) -> DeviceControlModel:
    """Rest end point for write device control."""
    energy_assistant = request.app.energy_assistant if hasattr(request.app, "energy_assistant") else None
    if energy_assistant is None:
        raise HTTPException(status_code=500)

    return await use_case.execute(disable_device_control, energy_assistant.home)


@router.get(
    "/device/{device_id}",
    response_model=ReadDeviceConfigResponse,
)
async def read(
    request: Request,
    device_id: Annotated[uuid.UUID, Path(..., description="")],
    use_case: Annotated[ReadDeviceConfiguration, Depends(ReadDeviceConfiguration)],
) -> ConfigModel:
    """REST end point for read a device configuration."""
    energy_assistant = request.app.energy_assistant if hasattr(request.app, "energy_assistant") else None
    if energy_assistant is None:
        raise HTTPException(status_code=500)
    return await use_case.execute(energy_assistant.config, device_id)


@router.put(
    "/device/{device_id}",
    response_model=ReadDeviceConfigResponse,
)
async def write(
    request: Request,
    data: dict,
    device_id: Annotated[uuid.UUID, Path(..., description="")],
    use_case: Annotated[WriteDeviceConfiguration, Depends(WriteDeviceConfiguration)],
) -> ConfigModel:
    """REST end point for read a device configuration."""
    energy_assistant = request.app.energy_assistant if hasattr(request.app, "energy_assistant") else None
    if energy_assistant is None:
        raise HTTPException(status_code=500)
    return await use_case.execute(energy_assistant.config, data, device_id, energy_assistant.home)
